@@ -84,6 +84,15 @@ const char B64Chars[64] = {
   '8', '9', '+', '/'
 };
 
+/* RFC4648 section 5 Base 64 Encoding with URL and Filename Safe Alphabet */
+const char B64Chars_urlsafe[64] = {
+  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
+  'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
+  'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
+  't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
+  '8', '9', '-', '_'
+};
+
 static void transform_to_7bit (BODY *a, FILE *fpin);
 
 static void encode_quoted (FGETCONV * fc, FILE *fout, int istext)
@@ -1383,8 +1392,8 @@ retry:
     else if ((WithCrypto & APPLICATION_SMIME) &&
              ((mutt_is_application_smime (hdr->content) & SMIMEENCRYPT) == SMIMEENCRYPT))
     {
-      chflags |= CH_MIME | CH_TXTPLAIN;
-      cmflags = MUTT_CM_DECODE | MUTT_CM_CHARCONV;
+      chflags |= CH_MIME | CH_NONEWLINE;
+      cmflags = MUTT_CM_DECODE_SMIME;
       pgp &= ~SMIMEENCRYPT;
     }
   }
@@ -1881,36 +1890,51 @@ static const char *find_word (const char *src)
 }
 
 /* like wcwidth(), but gets const char* not wchar_t* */
-static int my_width (const char *str, int col, int flags)
+static int my_width (const char *p, int col, int flags)
 {
   wchar_t wc;
   int l, w = 0, nl = 0;
-  const char *p = str;
+  size_t n, consumed;
+  mbstate_t mbstate;
 
-  while (p && *p)
+  if (!p)
+    return 0;
+
+  memset (&mbstate, 0, sizeof (mbstate));
+  n = mutt_strlen (p);
+
+  while (*p && n)
   {
-    if (mbtowc (&wc, p, MB_CUR_MAX) >= 0)
+    consumed = mbrtowc (&wc, p, n, &mbstate);
+    if (!consumed)
+      break;
+    if (consumed == (size_t)(-1) || consumed == (size_t)(-2))
     {
-      l = wcwidth (wc);
-      if (l < 0)
-	l = 1;
-      /* correctly calc tab stop, even for sending as the
-       * line should look pretty on the receiving end */
-      if (wc == L'\t' || (nl && wc == L' '))
-      {
-	nl = 0;
-	l = 8 - (col % 8);
-      }
-      /* track newlines for display-case: if we have a space
-       * after a newline, assume 8 spaces as for display we
-       * always tab-fold */
-      else if ((flags & CH_DISPLAY) && wc == '\n')
-	nl = 1;
+      if (consumed == (size_t)(-1))
+        memset (&mbstate, 0, sizeof (mbstate));
+      wc = replacement_char ();
+      consumed = (consumed == (size_t)(-1)) ? 1 : n;
     }
-    else
+
+    l = wcwidth (wc);
+    if (l < 0)
       l = 1;
+    /* correctly calc tab stop, even for sending as the
+     * line should look pretty on the receiving end */
+    if (wc == L'\t' || (nl && wc == L' '))
+    {
+      nl = 0;
+      l = 8 - (col % 8);
+    }
+    /* track newlines for display-case: if we have a space
+     * after a newline, assume 8 spaces as for display we
+     * always tab-fold */
+    else if ((flags & CH_DISPLAY) && wc == '\n')
+      nl = 1;
+
     w += l;
-    p++;
+    p += consumed;
+    n -= consumed;
   }
   return w;
 }
